/*
  k8056.c

  This is a control application for my Velleman K8056
  8-port RS232 relay card.

  Raimund Steger 2004-05-09, 2004-05-10
*/

#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#define DEFAULT_DEVICE "/dev/ttyS0"
#define DEFAULT_RETRIES 3
#define DEFAULT_RETRY_WAIT 100 /* ms */
#define DEFAULT_PULSE_WAIT 500 /* ms */
#define DEFAULT_TEST_CYCLE_WAIT 1000 /* ms */
#define DEFAULT_CARD_ADDRESS 1;

void show_usage() {
    printf("Usage:\n\n");
    printf("  k8056 [-d device] [-a address] (s|S) <relay_to_set>\n");
    printf("  k8056 [-d device] [-a address] (c|C) <relay_to_clear>\n");
    printf("  k8056 [-d device] [-a address] (t|T) <relay_to_toggle>\n");
    printf("  k8056 [-d device] [-a address] b <byte_to_send>\n");
    printf("  k8056 [-d device] [-a address] X <number_of_test_cycles>\n");
    printf("\nCommands s and c accept values from 1 to 9, where 9 means\n");
    printf("\"all relays\"; command t accepts values from 1 to 8; commands\n");
    printf("S, C and T correspond to their lower-case versions except that\n");
    printf("they revoke themselves after a short period of time; command\n");
    printf("b accepts values from 1 to 255, which is interpreted as bit\n");
    printf("mask used to set all relays at once. Command X sends an\n");
    printf("arbitrary number of testing cycles to the relay card.\n");
    printf("Defaults are: serial device /dev/ttyS0, card address 1.\n");
}

int is_relay_address(int all_possible,char *c) {
    if(all_possible)
	return
	    strcmp(c,"1")||strcmp(c,"2")||strcmp(c,"3")||
	    strcmp(c,"4")||strcmp(c,"5")||strcmp(c,"6")||
	    strcmp(c,"7")||strcmp(c,"8")||strcmp(c,"9");
    else
	return
	    strcmp(c,"1")||strcmp(c,"2")||strcmp(c,"3")||
	    strcmp(c,"4")||strcmp(c,"5")||strcmp(c,"6")||
	    strcmp(c,"7")||strcmp(c,"8");
}
	
unsigned char compute_checksum(unsigned char *c) {
    unsigned char sum=c[0]+c[1]+c[2]+c[3];
    return 256 - (sum & 0xFF);
}

void set_k8056(int fd,unsigned char *c) {
    int i;
    c[4]=compute_checksum(c);
    printf(
	"Setting %02x-%02x-%02x-%02x-%02x",
	c[0],c[1],c[2],c[3],c[4]);
    for(i=0;i<DEFAULT_RETRIES;i++) {
	if(i>0) usleep(DEFAULT_RETRY_WAIT*1000);
	write(fd,c,5);
	printf(".");
    }
    printf("\n");
}

int main(int argc,char *argv[]) {
    extern char *optarg;
    extern int optind;
    struct termios tio;
    int fd;
    char opt,*device,*cmd,*carg;
    unsigned char card_address,k8056cmd[5];

    /* setup variables & parse command line options */
    card_address=DEFAULT_CARD_ADDRESS;
    device=DEFAULT_DEVICE;
    while((opt=getopt(argc,argv,"a:d:"))!=-1) {
	switch(opt) {
	case 'a':
	    if(sscanf(optarg,"%i",&card_address)!=1) {
		fprintf(stderr,"Wrong address specified (1..255 possible).\n");
		exit(-1);
	    }
	    printf("Using card address %i.\n",card_address);
	    break;
	case 'd':
	    device=optarg;
	    printf("Using serial device %s.\n",device);
	    break;
	}
    }
    if((argc-optind)<2) {
	show_usage(); 
	exit(-1);
    }
    cmd=argv[optind];
    carg=argv[optind+1];

    /* setup serial communication */
    fd=open(device,O_RDWR|O_NOCTTY);
    if(fd<0) {
	fprintf(stderr,"Error opening %s.\n",device);
	exit(-1);
    }
    tio.c_cflag=CS8|CLOCAL;
    tio.c_iflag=IGNPAR;
    tio.c_oflag=0;
    tio.c_lflag=0;
    tio.c_cc[VMIN]=1;
    tio.c_cc[VTIME]=0;
    tcflush(fd,TCIOFLUSH);
    cfsetospeed(&tio,B2400);
    cfsetispeed(&tio,B2400);
    tcsetattr(fd,TCSANOW,&tio);
    
    /* this part of the command sequence is the same for all commands */
    k8056cmd[0]=13;
    k8056cmd[1]=card_address;

    /* interpret & execute command */
    /*   set a relay */
    if(strcmp(cmd,"s")==0) {
	if(is_relay_address(1,carg)) {
	    k8056cmd[2]='S';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'9'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   set (pulse) a relay */
    else if(strcmp(cmd,"S")==0) {
	if(is_relay_address(1,carg)) {
	    k8056cmd[2]='S';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	    usleep(DEFAULT_PULSE_WAIT*1000);
	    k8056cmd[2]='C';
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'9'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   clear a relay */
    else if(strcmp(cmd,"c")==0) {
	if(is_relay_address(1,carg)) {
	    k8056cmd[2]='C';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'9'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   clear (pulse) a relay */
    else if(strcmp(cmd,"C")==0) {
	if(is_relay_address(1,carg)) {
	    k8056cmd[2]='C';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	    usleep(DEFAULT_PULSE_WAIT*1000);
	    k8056cmd[2]='S';
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'9'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   toggle a relay */
    else if(strcmp(cmd,"t")==0) {
	if(is_relay_address(0,carg)) {
	    k8056cmd[2]='T';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'8'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   toggle (pulse) a relay */
    else if(strcmp(cmd,"T")==0) {
	if(is_relay_address(0,carg)) {
	    k8056cmd[2]='T';
	    k8056cmd[3]=carg[0];
	    set_k8056(fd,(char *)&k8056cmd);
	    usleep(DEFAULT_PULSE_WAIT*1000);
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay # '1'..'8'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   change all relays */
    else if(strcmp(cmd,"b")==0) {
	unsigned int relay_mask;
	if(sscanf(carg,"%i",&relay_mask)==1) {
	    k8056cmd[2]='B';
	    k8056cmd[3]=(char)(relay_mask & 0xFF);
	    set_k8056(fd,(char *)&k8056cmd);
	}
	else {
	    fprintf(stderr,"Must specify relay mask 0x0..0xff'.\n");
	    close(fd);
	    exit(-1);
	}
    }
    /*   test relays */
    else if(strcmp(cmd,"X")==0) {
	unsigned int num_cycles;
	unsigned int relay_mask=1;
	int i;
	if(sscanf(carg,"%i",&num_cycles)==1) {
	    k8056cmd[2]='B';
	    for(i=0;i<num_cycles;i++) {
		if(i>0) usleep(DEFAULT_TEST_CYCLE_WAIT*1000);
		k8056cmd[3]=(char)(relay_mask & 0xFF);
		relay_mask=(relay_mask==0x80)?1:(relay_mask<<1);
		set_k8056(fd,(char *)&k8056cmd);
	    }
	}
	else {
	    fprintf(stderr,"Must specify number of testing cycles.\n");
	    close(fd);
	    exit(-1);
	}
    }
    else {
	fprintf(stderr,"k8056: %s: Unknown command.\n",cmd);
	show_usage();
	close(fd);
	exit(-1);
    }
    close(fd);
    exit(0);
}
